library mhc_logging;

/// The singleton logger for the current isolate.
final logger = new Logger._();

/// Logs with the [Level.DEBUG] level.
void debug(String code, [argument, String domain])
  => logger.log(Level.DEBUG, code, argument, domain);

/// Logs with the [Level.INFO] level.
void info(String code, [argument, String domain])
  => logger.log(Level.INFO, code, argument, domain);

/// Logs with the [Level.CONFIG] level.
void config(String code, [argument, String domain])
  => logger.log(Level.CONFIG, code, argument, domain);

/// Logs with the [Level.OK] level.
void ok(String code, [argument, String domain])
  => logger.log(Level.OK, code, argument, domain);

/// Logs with the [Level.WARNING] level.
void warn(String code, [argument, String domain])
  => logger.log(Level.WARNING, code, argument, domain);

/// Logs with the [Level.ERROR] level.
void error(String code, [argument, String domain])
  => logger.log(Level.ERROR, code, argument, domain);

/// Logs with the [Level.INTERNAL] level.
void internal(String code, [argument, String domain])
  => logger.log(Level.INTERNAL, code, argument, domain);

/// Logs with the [Level.CRITICAL] level.
void critical(String code, [argument, String domain])
  => logger.log(Level.CRITICAL, code, argument, domain);

class Level {
  static const int ALL      =    0;
  static const int DEBUG    =  100;
  static const int INFO     =  200;
  static const int CONFIG   =  300;
  static const int OK       =  400;
  static const int WARNING  =  500;
  static const int ERROR    =  600;
  static const int INTERNAL =  700;
  static const int CRITICAL =  800;
  static const int OFF      = 1000;
}

class LogRecord {
  final String    code;
  final int       level;
  final           argument;
  final DateTime  date;
  final String    domain;
  
  LogRecord(this.code, this.level, this.argument, this.date, this.domain);
  
  toJson() {
    var result = {'code': code, 'level': level, 'date': date};
    if (argument != null) result['argument'] = argument;
    if (domain != null) result['domain'] = domain;
    return result;
  }
  
  toString() {
    var result = 'LogRecord(';
    result += '\n\tcode = "$code",';
    result += '\n\tlevel = $level,';
    result += '\n\tdate = "${date.toString()}",';
    if (argument != null) result += '\n\targument = $argument,';
    if (domain != null) result += '\n\tdomain = "$domain",';
    return result + '\n)';
  }
}

class Logger {
  List<LogListener> _listeners;
  
  Logger._();
  
  void addListener(LogListener listener) {
    if (_listeners == null) {
      _listeners = new List();
      _listeners.add(listener);
    } else {
      if (! _listeners.contains(listener)) _listeners.add(listener);
    }
  }
  
  void removeListener(LogListener listener) {
    if (_listeners != null) _listeners.remove(listener);
  }
  
  void clearListeners() {
    if (_listeners != null) _listeners.clear();
  }
  
  void log(int level, String code, [argument, String domain]) {
    if (_listeners != null) {
      LogRecord log;
      for (var listener in _listeners) {
        if (listener.isLoggable(level, domain)) {
          if (log == null) log = new LogRecord(code, level, argument, new DateTime.now(), domain);
          listener.handle(log);
        }
      }
    }
  }
}

abstract class LogListener {
  final int     level;
  final String  name;
  final Pattern code;
  final Pattern domain;
  
  const LogListener(this.level, {this.name, this.code, this.domain});
  
  bool isLoggable(int level, [String code, String domain]) =>
      level >= this.level
        && (this.code == null || this.code.allMatches(code).length > 0)
        && (this.domain == null || this.domain.allMatches(domain).length > 0);
  
  void handle(LogRecord log);
}

class LogPrinter extends LogListener {
  const LogPrinter(int level, {String name, String code, String domain}):
      super(level, name: name, code: code, domain: domain);
  
  void handle(LogRecord log) => print(log.toString());
}
